/*
 * main.cpp
 * 
 * Copyright 2012 Alexander Andrejevic <theflash@sdf.lonestar.org>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 */

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cstdarg>
#include <cmath>
#include <ctime>
#include "asteroid.h"
#include "model3D.h"
#include "tga_loader.h"

#ifdef WIN32
	#include <windows.h>
#else
	#include <unistd.h>
#endif

#define MAX_LIVES 3
#define NUM_TEXTURES 7

using namespace std;

typedef enum
{
	GAME_INTRO_SCREEN,
	GAME_PLAYING,
	GAME_SPACESHIP_HIT,
	GAME_PAUSED,
	GAME_OVER
} CURRENT_GAME_STATUS;

typedef struct
{
	int timer;
	GLdouble x, y, z;
	GLfloat power;
	int id;
} LASER;

typedef struct
{
	bool active;
	GLfloat z, power;
} PHOTONIC_CANNON;

const char *img_file_list[NUM_TEXTURES]=
{
	"textures/intro.tga",
	"textures/space.tga",
	"textures/viewscreen.tga",
	"textures/viewscreen_mask.tga",
	"textures/star.tga",
	"textures/hollow_star.tga",
	"textures/explosion.tga"
};

bool cheat=false;
CURRENT_GAME_STATUS status;
int g_width, g_height;
clock_t ticks=0;
GLuint textures[NUM_TEXTURES];
vector <asteroid*> asteroids;
model3D *asteroid_model;
int level=1, lives=MAX_LIVES, level_waves, last_update;
bool level_start=true, g_sounds=true;
LASER laser;
PHOTONIC_CANNON cannon;
GLUquadric *quadric;
int flash_counter, ast_dest=0, num_hits=0, num_misses=0;

void glprintf(void *font, const char *fmt, ...)
{
	char buffer[256];
    va_list vl;
    va_start(vl, fmt);
	vsprintf(buffer, fmt, vl);
	va_end(vl);
	for(int i=0; buffer[i]!=0; i++) glutBitmapCharacter(font, buffer[i]);
}

void play_sound(const char *file_path)
{
	if(!g_sounds) return;
#ifdef WIN32
	PlaySoundA(file_path, NULL, SND_FILENAME | SND_ASYNC);
#else
	if(fork()==0)
	{
		freopen("/dev/null", "r", stdin);
		freopen("/dev/null", "w", stdout);
		freopen("/dev/null", "w", stderr);
		execlp("/usr/bin/play", "play", file_path, NULL);
		exit(0);
	}
#endif
}

void reshape(GLint width, GLint height)
{
	g_width=width;
	g_height=height;
	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 100.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void load_textures()
{
	glEnable(GL_TEXTURE_2D);
	for(int i=0; i<NUM_TEXTURES; i++)
	{
	    tga_image *image=new tga_image(img_file_list[i]);
	    glGenTextures(1, &textures[i]);
	    glBindTexture(GL_TEXTURE_2D, textures[i]);
		glTexImage2D(GL_TEXTURE_2D, 0, image->bpp/8, image->width, image->height, 0, (image->bpp==24) ? GL_RGB:GL_RGBA, GL_UNSIGNED_BYTE, image->data);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		delete image;
	}
	glDisable(GL_TEXTURE_2D);
}

void fire_cannon()
{
	if(cannon.power<100.0) return;
	play_sound("sounds/cannon.wav");
	cannon.z=-1.0f;
	cannon.active=true;
}

void fire_laser(int mouse_x, int mouse_y)
{
	if(laser.power<100) return;
	play_sound("sounds/laser.wav");
	laser.id=-1;
	laser.power=0;
	GLuint pick_buffer[1024];
	GLint viewport[4];
	GLdouble modelview_matrix[16], projection_matrix[16];
	glGetIntegerv(GL_VIEWPORT, viewport);
	glSelectBuffer(1024*sizeof(GLint), pick_buffer);
	glRenderMode(GL_SELECT);
	glInitNames();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluPickMatrix(mouse_x, g_height-mouse_y, 0.1f, 0.1f, viewport);
	gluPerspective(45.0f, (GLfloat)((GLfloat)viewport[2]-(GLfloat)viewport[0])/(GLfloat)((GLfloat)viewport[3]-(GLfloat)viewport[1]), 0.1f, 100.0f);
	glMatrixMode(GL_MODELVIEW);
	for(int i=0; i<asteroids.size(); i++)
	{
		glPushName(i);
		asteroids[i]->draw();
		glPopName();
	}
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);	
	GLint selected=glRenderMode(GL_RENDER);
	GLfloat mouse_z;
	glReadPixels(mouse_x, g_height-mouse_y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &mouse_z);
	glGetDoublev(GL_PROJECTION_MATRIX, projection_matrix);
	glGetDoublev(GL_MODELVIEW_MATRIX, modelview_matrix);
	gluUnProject((GLdouble)mouse_x, (GLdouble)(g_height-mouse_y), mouse_z, modelview_matrix, projection_matrix, viewport, &laser.x, &laser.y, &laser.z);
	if(selected>0)
	{
		int sel=pick_buffer[3];
		int min_z=pick_buffer[1];
		for(int i=0; i<selected; i++) if(pick_buffer[i*4+1]<min_z)
		{
		    sel=pick_buffer[i*4+3];
			min_z=pick_buffer[i*4+1];
		}
		laser.id=sel;
		float laser_power;
		laser_power=-750.0f/asteroids[sel]->z;
		if(!asteroids[sel]->explosion && asteroids[sel]->exists)
		{
			if(asteroids[sel]->laser_hit(laser_power))
		    {
			    asteroids[sel]->explode();
			    ast_dest++;
		    }
			num_hits++;
		}
		else num_misses++;
	}
	else num_misses++;
	laser.timer=300;
}

void draw_hud()
{
	glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT | GL_CURRENT_BIT | GL_TEXTURE_BIT);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, (GLdouble)g_width, (GLdouble)g_height, 0, -1.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glBlendFunc(GL_DST_COLOR, GL_ZERO);
	glBindTexture(GL_TEXTURE_2D, textures[3]);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0f, 0.0f); glVertex2f(0.0f, (GLfloat)g_height);
		glTexCoord2f(1.0f, 0.0f); glVertex2f((GLfloat)g_width, (GLfloat)g_height);
		glTexCoord2f(1.0f, 1.0f); glVertex2f((GLfloat)g_width, 0.0f);
		glTexCoord2f(0.0f, 1.0f); glVertex2f(0.0f, 0.0f);
	glEnd();
	glBlendFunc(GL_ONE, GL_ONE);
	glBindTexture(GL_TEXTURE_2D, textures[2]);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0f, 0.0f); glVertex2f(0.0f, (GLfloat)g_height);
		glTexCoord2f(1.0f, 0.0f); glVertex2f((GLfloat)g_width, (GLfloat)g_height);
		glTexCoord2f(1.0f, 1.0f); glVertex2f((GLfloat)g_width, 0.0f);
		glTexCoord2f(0.0f, 1.0f); glVertex2f(0.0f, 0.0f);
	glEnd();
	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);
	glColor3f(0.0f, 0.0f, 0.0f);
	glRasterPos2f((GLfloat)g_width/32.0f, (GLfloat)g_height/24.0f);
	glprintf(GLUT_BITMAP_HELVETICA_18, "Level: %d", level);
	glRasterPos2f((GLfloat)g_width/4.0f, (GLfloat)g_height/24.0f);
	glprintf(GLUT_BITMAP_HELVETICA_18, "Lives: %d", lives);
	glRasterPos2f((4.0f*(GLfloat)g_width)/9.0f, (GLfloat)g_height/24.0f);
	glprintf(GLUT_BITMAP_HELVETICA_18, "Laser power:");
	if(laser.power==100) glColor3f(0.0f, 0.8f, 0.0f);
	else glColor3f(0.8f, 0.0f, 0.0f);
	glRectf((3.0*((GLfloat)g_width))/5.0f, (GLfloat)g_height/60.0f, (3.0*((GLfloat)g_width))/5.0f+(GLfloat)(laser.power)*((GLfloat)g_width/300.0f), (GLfloat)g_height/20.0f);
	glColor3f(0.0f, 0.0f, 0.0f);
	glRasterPos2f((GLfloat)g_width/32.0f, 18.0f+(23.0f*(GLfloat)g_height)/24.0f);
	glprintf(GLUT_BITMAP_HELVETICA_18, "Photonic cannon: ");
	if(cannon.power==100) glColor3f(0.0f, 0.8f, 0.0f);
	else if(!cannon.active) glColor3f(0.8f, 0.0f, 0.0f);
	else glColor3f(0.0f, 0.0f, 0.8f);
	glRectf(((GLfloat)g_width)/4.0f, (23.0f*(GLfloat)g_height)/24.0f, ((GLfloat)g_width)/4.0f+(GLfloat)(cannon.power)*((GLfloat)g_width/200.0f), (119.0f*(GLfloat)g_height/120.0f));
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0f, (GLfloat)g_width/(GLfloat)g_height, 0.1f, 100.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
    glPopAttrib();
}

void display()
{
	int text_width=0, score=0, star=0, hollow_star=0;
	GLfloat current_y;
	char rank[32];
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	switch(status)
	{
	case GAME_INTRO_SCREEN:
		glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT | GL_CURRENT_BIT | GL_TEXTURE_BIT);
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_DEPTH_TEST);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(0.0, (GLdouble)g_width, (GLdouble)g_height, 0, -1.0, 1.0);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glBindTexture(GL_TEXTURE_2D, textures[0]);
		glBegin(GL_QUADS);
			glTexCoord2f(0.0f, 0.0f); glVertex2f(0.0f, (GLfloat)g_height);
			glTexCoord2f(1.0f, 0.0f); glVertex2f((GLfloat)g_width, (GLfloat)g_height);
			glTexCoord2f(1.0f, 1.0f); glVertex2f((GLfloat)g_width, 0.0f);
			glTexCoord2f(0.0f, 1.0f); glVertex2f(0.0f, 0.0f);
		glEnd();
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(45.0f, (GLfloat)g_width/(GLfloat)g_height, 0.1f, 100.0f);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glPopAttrib();
	    break;
	case GAME_PLAYING:
		glEnable(GL_TEXTURE_2D);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glBindTexture(GL_TEXTURE_2D, textures[1]);
		glBegin(GL_QUADS);
  			glTexCoord2f(1.0f, 1.0f); glVertex3f(55.0f, 41.5f, -100.0f);
			glTexCoord2f(0.0f, 1.0f); glVertex3f(-55.0f, 41.5f, -100.0f);
			glTexCoord2f(0.0f, 0.0f); glVertex3f(-55.0f, -41.5f, -100.0f);
			glTexCoord2f(1.0f, 0.0f); glVertex3f(55.0f, -41.5f, -100.0f);
		glEnd();
		glDisable(GL_TEXTURE_2D);
		for(int i=0; i<asteroids.size(); i++) asteroids[i]->draw();
		if(laser.timer>0)
		{
			glLoadIdentity();
	        glColor3f(1.0f, 0.0f, 0.0f);
	        glTranslatef(0.0f, -1.0f, -1.0f);
	        if(laser.id!=-1) laser.z=(GLfloat)asteroids[laser.id]->z;
	        GLfloat dx=(GLfloat)laser.x, dy=(GLfloat)laser.y+1.0f, dz=laser.z+1.0f;
	        GLfloat length=sqrt(dx*dx+dy*dy+dz*dz);
			GLfloat angle=(180.0/PI)*acos(dz/length);
			if(dz<0.0f) angle=-angle;
			glRotatef(angle, -dy*dz, dx*dz, 0.0f);
	        gluCylinder(quadric, 0.05, 0.05, length, 100, 100);
		    glColor3f(1.0f, 1.0f, 1.0f);
		}
		if(cannon.active)
		{
			glLoadIdentity();
			glTranslatef(0.0f, 0.0f, cannon.z);
			glColor3f(0.0f, 0.5f, 1.0f);
			glutSolidTorus(0.6f, 3.0f, 100, 100);
			glColor3f(1.0f, 1.0f, 1.0f);
		}
		draw_hud();
		break;
	case GAME_SPACESHIP_HIT:
		glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT | GL_CURRENT_BIT);
		glDisable(GL_DEPTH_TEST);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(0.0, (GLdouble)g_width, (GLdouble)g_height, 0, -1.0, 1.0);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		if((flash_counter/100)%2==0) glColor3f(1.0f, 0.0f, 0.0f);
		else glColor3f(1.0f, 1.0f, 0.0f);
		glBegin(GL_QUADS);
			glVertex2f(0.0f, (GLfloat)g_height);
			glVertex2f((GLfloat)g_width, (GLfloat)g_height);
			glVertex2f((GLfloat)g_width, 0.0f);
			glVertex2f(0.0f, 0.0f);
		glEnd();
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(45.0f, (GLfloat)g_width/(GLfloat)g_height, 0.1f, 100.0f);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glPopAttrib();
		break;
	case GAME_PAUSED:
		glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT | GL_CURRENT_BIT);
		glDisable(GL_DEPTH_TEST);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(0.0, (GLdouble)g_width, (GLdouble)g_height, 0, -1.0, 1.0);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glColor3f(1.0f, 1.0f, 1.0f);
		text_width=glutBitmapLength(GLUT_BITMAP_HELVETICA_18, (const unsigned char*)"Game paused - press space to continue");
		glRasterPos2f(((GLfloat)g_width-text_width)/2.0f, (GLfloat)g_height/2.0f);
		glprintf(GLUT_BITMAP_HELVETICA_18, "Game paused - press space to continue");
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(45.0f, (GLfloat)g_width/(GLfloat)g_height, 0.1f, 100.0f);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glPopAttrib();
		break;
	case GAME_OVER:
		glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT | GL_CURRENT_BIT);
		glDisable(GL_DEPTH_TEST);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(0.0, (GLdouble)g_width, (GLdouble)g_height, 0, -1.0, 1.0);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glColor3f(1.0f, 1.0f, 1.0f);
		current_y=(GLfloat)(g_height-390.0f)/2.0f;
		text_width=glutBitmapLength(GLUT_BITMAP_HELVETICA_18, (const unsigned char*)"GAME OVER");
		glRasterPos2f(((GLfloat)g_width-text_width)/2.0f, current_y);
		glprintf(GLUT_BITMAP_HELVETICA_18, "GAME OVER");
		current_y+=60.0f;
		text_width=glutBitmapLength(GLUT_BITMAP_HELVETICA_18, (const unsigned char*)"Your score:");
		glRasterPos2f(((GLfloat)g_width-text_width)/2.0f, current_y);
		glprintf(GLUT_BITMAP_HELVETICA_18, "Your score:");
		current_y+=60.0f;
		glRasterPos2f(((GLfloat)g_width-450)/2.0f, current_y);
		glprintf(GLUT_BITMAP_9_BY_15, "Total asteroids destroyed: %4d    Score: %8d", ast_dest, ast_dest*100);
		score+=ast_dest*100;
		current_y+=20.0f;
		glRasterPos2f((GLfloat)(g_width-450)/2.0f, current_y);
		if((num_hits+num_misses)!=0)
		{
			int percentage=(int)((float)num_hits*100.0f/(float)(num_hits+num_misses));
			glprintf(GLUT_BITMAP_9_BY_15, "Shot accuracy:             %3d%%    Score: %8d", percentage, percentage*100);
			score+=percentage*100;
		}
		else glprintf(GLUT_BITMAP_9_BY_15, "Shot accuracy:              N/A    Score:        0");
		current_y+=20.0f;
		glRasterPos2f((GLfloat)(g_width-450)/2.0f, current_y);
		glprintf(GLUT_BITMAP_9_BY_15, "Lives lost:                %4d    Score: %8d", MAX_LIVES-lives, (lives-MAX_LIVES)*1000);
		score+=(lives-MAX_LIVES)*1000;
		current_y+=20.0f;
		glRasterPos2f((GLfloat)(g_width-450)/2.0f, current_y);
		glprintf(GLUT_BITMAP_9_BY_15, "Total levels won:          %4d    Score: %8d", level-1, ((level*(level-1))/2)*1000);
		score+=((level*(level-1))/2)*1000;
		current_y+=20.0f;
		glRasterPos2f((GLfloat)(g_width-450)/2.0f, current_y);
		glprintf(GLUT_BITMAP_9_BY_15, "TOTAL SCORE:                              %8d", score);
		current_y+=60.0f;
		text_width=glutBitmapLength(GLUT_BITMAP_HELVETICA_18, (const unsigned char*)"Which gives you the rank of:");
		glRasterPos2f((GLfloat)(g_width-text_width)/2.0f, current_y);
		glprintf(GLUT_BITMAP_HELVETICA_18, "Which gives you the rank of:");
		current_y+=40.0f;
		if(score>=81000)
		{
			strcpy(rank, "ADMIRAL");
			star=5;
		}
		else if(score>=67000 && score<81000)
		{
			strcpy(rank, "Captain");
			star=4;
		}
		else if(score>=54400 && score<67000)
		{
			strcpy(rank, "Commander");
			star=3;
		}
		else if(score>=43200 && score<54400)
		{ 
			strcpy(rank, "Lieutenant Commander");
			star=2;
			hollow_star=1;
		}
        else if(score>=25000 && score<43200)
		{
		    strcpy(rank, "Lieutenant");
			star=2;
		}
        else if(score>=14200 && score<25000)
		{
			strcpy(rank, "Lieutenant J.G.");
			star=1;
			hollow_star=1;
		}
        else if(score<14200)
		{
		    strcpy(rank, "Ensign");
			star=1;
		}
		text_width=glutBitmapLength(GLUT_BITMAP_TIMES_ROMAN_24, (const unsigned char*)rank);
		glRasterPos2f((GLfloat)(g_width-text_width)/2.0f, current_y);
		glprintf(GLUT_BITMAP_TIMES_ROMAN_24, rank);
		current_y+=40.0f;
		for(int i=0; i<star; i++)
		{
			GLfloat center=(g_width-((star+hollow_star)*50.0f))/2.0f;
			glEnable(GL_TEXTURE_2D);
			glTranslatef(center+(GLfloat)i*50.0f, current_y, 0.0f);
			glBindTexture(GL_TEXTURE_2D, textures[4]);
			glBegin(GL_QUADS);
				glTexCoord2f(0.0f, 0.0f); glVertex2f(0.0f, 50.0f);
				glTexCoord2f(1.0f, 0.0f); glVertex2f(50.0f, 50.0f);
				glTexCoord2f(1.0f, 1.0f); glVertex2f(50.0f, 0.0f);
				glTexCoord2f(0.0f, 1.0f); glVertex2f(0.0f, 0.0f);
			glEnd();
			glLoadIdentity();
			glDisable(GL_TEXTURE_2D);
		}
		for(int i=0; i<hollow_star; i++)
		{
			GLfloat center=(g_width-((star+hollow_star)*50.0f))/2.0f;
			glEnable(GL_TEXTURE_2D);
			glTranslatef(center+(GLfloat)(i+star)*50.0f, current_y, 0.0f);
			glBindTexture(GL_TEXTURE_2D, textures[5]);
			glBegin(GL_QUADS);
				glTexCoord2f(0.0f, 0.0f); glVertex2f(0.0f, 50.0f);
				glTexCoord2f(1.0f, 0.0f); glVertex2f(50.0f, 50.0f);
				glTexCoord2f(1.0f, 1.0f); glVertex2f(50.0f, 0.0f);
				glTexCoord2f(0.0f, 1.0f); glVertex2f(0.0f, 0.0f);
			glEnd();
			glLoadIdentity();
			glDisable(GL_TEXTURE_2D);
		}
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(45.0f, (GLfloat)g_width/(GLfloat)g_height, 0.1f, 100.0f);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glPopAttrib();
		break;
	}
	glutSwapBuffers();
}

void keyboard(unsigned char key, int x, int y)
{
	if(key==' ')
	{
		if(status==GAME_INTRO_SCREEN || status==GAME_PAUSED) status=GAME_PLAYING;
		else if(status==GAME_OVER)
		{
			asteroids.clear();
			lives=MAX_LIVES;
			laser.timer=0;
			cannon.power=0.0f;
			ast_dest=0;
			num_hits=0;
			num_misses=0;
			level=1;
			level_start=true;
			cheat=false;
			status=GAME_INTRO_SCREEN;
		}
		else if(status==GAME_PLAYING) fire_laser(x, y);
	}
	else if(key=='c' || key=='C')
	{
		if(status==GAME_PLAYING) fire_cannon();
	}
	else if(key==0x1B)
	{
		if(status==GAME_PLAYING || status==GAME_SPACESHIP_HIT || status==GAME_PAUSED) status=GAME_PAUSED;
		else exit(0);
    }
}

void special(int key, int x, int y)
{
	if(key==GLUT_KEY_F12 && (status==GAME_PLAYING || status==GAME_SPACESHIP_HIT || status==GAME_PAUSED)) status=GAME_OVER;
	if((glutGetModifiers()&GLUT_ACTIVE_SHIFT) && key==GLUT_KEY_F11 && (status==GAME_PLAYING || status==GAME_SPACESHIP_HIT || status==GAME_PAUSED)) cheat=!cheat;
}

void mouse(int button, int state, int x, int y)
{
	if(button==GLUT_LEFT_BUTTON && state==GLUT_DOWN) fire_laser(x, y);
}

void update(clock_t time)
{
	bool clean;
	switch(status)
    {
	case GAME_PLAYING:
		if(level_start)
		{
			level_waves=2*level;
			level_start=false;
		}
		if(asteroids.size()==0)
		{
			if(level_waves)
			{
				asteroids.push_back(new asteroid(level, asteroid_model, textures[6]));
				asteroids.push_back(new asteroid(level, asteroid_model, textures[6]));
				asteroids[1]->z=-65.0f;
				level_waves--;
			}
			else
			{
				if(lives>0) level++;
				level_start=true;
			}
		}
		clean=false;
		while(!clean)
		{
			clean=true;
			for(int i=0; i<asteroids.size(); i++) if(!asteroids[i]->exists)
			{
				asteroids.erase(asteroids.begin()+i);
				clean=false;
				break;
			}
		}
		if(laser.timer>0) laser.timer-=time;
		if(cannon.active)
		{
			cannon.z-=(GLfloat)time/10.0f;
			if(cannon.z<=-100.0f)
			{
				play_sound("sounds/cannon.wav");
				cannon.z=-1.0f;
			}
			for(int i=0; i<asteroids.size(); i++) if(!asteroids[i]->explosion)
			{
				if(asteroids[i]->laser_hit((GLfloat)time/20.0f))
				{
					asteroids[i]->explode();
					ast_dest++;
				}
			}
			cannon.power-=0.005f*(GLfloat)time;
			if(cannon.power<=0.0f)
			{
				cannon.active=false;
				cannon.power=0.0f;
			}
		}
		else
		{
			if(cannon.power<100.0f) cannon.power+=0.002f*(GLfloat)time;
			if(cannon.power>100.0f) cannon.power=100.0f;
		}
		if(laser.power<100.0f) laser.power+=0.2f*(GLfloat)time;
		if(laser.power>100.0f) laser.power=100.0f;
		if(lives<=0 || level>10) status=GAME_OVER;
		for(int i=0; i<asteroids.size(); i++)
		{
		    if(!asteroids[i]->move(time))
			{
				play_sound("sounds/explosion.wav");
				if(!cheat) lives--;
				flash_counter=1000;
				status=GAME_SPACESHIP_HIT;
				cannon.power=0.0f;
				asteroids.erase(asteroids.begin()+i);
			}
		}
		break;
	case GAME_SPACESHIP_HIT:
		flash_counter-=time;
		if(flash_counter<=0) status=GAME_PLAYING;
		break;
	}
}

void idle()
{
	int current_time=glutGet(GLUT_ELAPSED_TIME);
	if((current_time-last_update)>=20)
	{
		update(current_time-last_update);
		last_update=current_time;
		glutPostRedisplay();
	}
	else
	{
#ifndef WIN32
		usleep(10000);
#else
		Sleep(10);
#endif
	}
}

int main(int argc, char *argv[])
{
	bool fullscreen=false;
	int width=800, height=600;
	glutInit(&argc, argv);
	FILE *fd=fopen("fasteroids.conf", "r");
	if(fd!=NULL)
	{
		char buffer[256];
		while(!feof(fd))
		{
			memset(buffer, 0, sizeof(buffer));
			fgets(buffer, sizeof(buffer), fd);
			for(int i=0; buffer[i]!=0; i++)
			{
				if(buffer[i]=='\n') buffer[i]=0;
				else if(buffer[i]>='A' && buffer[i]<='Z') buffer[i]+=32;
			}
			char *split=strchr(buffer, '=');
			if(split!=NULL)
			{
				if(strncmp(buffer, "resolution", (int)(split-buffer))==0)
				{
					sscanf(split+1, "%dx%d", &width, &height);
					if(width<640) width=640;
					if(height<480) height=480;
				}
				else if(strncmp(buffer, "fullscreen", (int)(split-buffer))==0)
				{
					if(strcmp(split+1, "yes")==0 || strcmp(split+1, "true")==0) fullscreen=true;
				}
				else if(strncmp(buffer, "sounds", (int)(split-buffer))==0)
				{
					if(strcmp(split+1, "no")==0 || strcmp(split+1, "false")==0) g_sounds=false;
				}
			}
		}
		fclose(fd);
	}
	glutInitWindowSize(width, height);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
	glutCreateWindow("3D Fasteroids!");
	if(fullscreen) glutFullScreen();
	glShadeModel(GL_SMOOTH);
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClearDepth(1.0f);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glEnable(GL_LINE_SMOOTH);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	quadric=gluNewQuadric();
	load_textures();
	asteroid_model=new model3D("models/asteroid.obj");
	status=GAME_INTRO_SCREEN;
	glutReshapeFunc(reshape);
	glutDisplayFunc(display);
	glutKeyboardFunc(keyboard);
	glutSpecialFunc(special);
	glutMouseFunc(mouse);
	glutIdleFunc(idle);
	last_update=glutGet(GLUT_ELAPSED_TIME);
	glutMainLoop();
	return 0;
}
